@@ -1,11 +1,13 @@
 #
 # Copyright (C) 2019 FreeIPA Contributors see COPYING for license
 #
+from __future__ import division
 
 from datetime import datetime, timezone, timedelta
 import itertools
 import logging
 import os
+import socket
 import tempfile
 
 from ipahealthcheck.ipa.plugin import IPAPlugin, registry
@@ -17,13 +19,12 @@ from ipalib import api
 from ipalib import errors
 from ipalib import x509
 from ipalib.install import certmonger
-from ipalib.constants import RENEWAL_CA_NAME
+from ipalib.constants import RENEWAL_CA_NAME, IPA_CA_RECORD
 from ipaplatform.paths import paths
 from ipaserver.install import certs
 from ipaserver.install import dsinstance
 from ipaserver.install import krainstance
 from ipaserver.install import krbinstance
-from ipaserver.install import installutils
 from ipaserver.plugins import ldap2
 from ipapython import certdb
 from ipapython import ipautil
@@ -208,14 +209,14 @@ class IPACertmongerExpirationCheck(IPAPlugin):
             else:
                 delta = nafter - now
                 diff = int(delta.total_seconds() / DAY)
-                if diff < self.config.cert_expiration_days:
+                if diff < int(self.config.cert_expiration_days):
                     yield Result(self, constants.WARNING,
                                  key=id,
                                  expiration_date=generalized_time(nafter),
                                  days=diff,
                                  msg='Request id {key} expires in {days} '
                                      'days. certmonger should renew this '
-                                     'automatically. Watch the status with'
+                                     'automatically. Watch the status with '
                                      'getcert list -i {key}.')
                 else:
                     yield Result(self, constants.SUCCESS,
@@ -309,7 +310,7 @@ class IPACertfileExpirationCheck(IPAPlugin):
 
             delta = notafter - now
             diff = int(delta.total_seconds() / DAY)
-            if diff < self.config.cert_expiration_days:
+            if diff < int(self.config.cert_expiration_days):
                 yield Result(self, constants.WARNING,
                              key=id,
                              expiration_date=generalized_time(notafter),
@@ -394,6 +395,111 @@ class IPACertTracking(IPAPlugin):
                                  'is not expected on an IPA master.')
 
 
+@registry
+class IPACertDNSSAN(IPAPlugin):
+    """Check whether a IPA-issued certificates have a SAN configured
+
+       Steps:
+       1. Collect all expected certificates into `requests`
+       2. Iterate over the list of certificates
+       3. If issued by IPA and a caIPAserviceCert then verify that
+          the host FQDN is in the list of SAN
+       4. If a CA is configured on this host then also verify that
+          ipa-ca.$DOMAIN is in the SAN.
+    """
+
+    requires = ('dirsrv',)
+
+    @duration
+    def check(self):
+        fqdn = socket.getfqdn()
+        requests = get_expected_requests(self.ca, self.ds, self.serverid)
+
+        for request in requests:
+            request_id = certmonger.get_request_id(request)
+            if request_id is None:
+                yield Result(self, constants.ERROR,
+                             key=request_id,
+                             msg='Found request id {key} but it is not tracked'
+                                 'by certmonger!?')
+                continue
+
+            ca_name = certmonger.get_request_value(request_id, 'ca-name')
+            if ca_name != 'IPA':
+                logger.debug('Skipping request %s with CA %s',
+                             request_id, ca_name)
+                continue
+            profile = certmonger.get_request_value(request_id,
+                                                   'template_profile')
+            if profile != 'caIPAserviceCert':
+                logger.debug('Skipping request %s with profile %s',
+                             request_id, profile)
+                continue
+
+            certfile = None
+            if request.get('cert-file') is not None:
+                certfile = request.get('cert-file')
+                try:
+                    cert = x509.load_certificate_from_file(certfile)
+                except Exception as e:
+                    yield Result(self, constants.ERROR,
+                                 key=request_id,
+                                 certfile=certfile,
+                                 error=str(e),
+                                 msg='Unable to open cert file {certfile}: '
+                                     '{error}')
+                    continue
+            elif request.get('cert-database') is not None:
+                nickname = request.get('cert-nickname')
+                dbdir = request.get('cert-database')
+                try:
+                    db = certdb.NSSDatabase(dbdir)
+                except Exception as e:
+                    yield Result(self, constants.ERROR,
+                                 key=request_id,
+                                 dbdir=dbdir,
+                                 error=str(e),
+                                 msg='Unable to open NSS database {dbdir}: '
+                                     '{error}')
+                    continue
+                try:
+                    cert = db.get_cert(nickname)
+                except Exception as e:
+                    yield Result(self, constants.ERROR,
+                                 key=id,
+                                 dbdir=dbdir,
+                                 nickname=nickname,
+                                 error=str(e),
+                                 msg='Unable to retrieve certificate '
+                                     '\'{nickname}\' from {dbdir}: {error}')
+                    continue
+
+            hostlist = [fqdn]
+            if self.ca.is_configured() and certfile == paths.HTTPD_CERT_FILE:
+                hostlist.append(f'{IPA_CA_RECORD}.{api.env.domain}')
+            error = False
+            for host in hostlist:
+                if host not in cert.san_a_label_dns_names:
+                    error = True
+                    yield Result(self, constants.ERROR,
+                                 key=request_id,
+                                 hostname=host,
+                                 san=cert.san_a_label_dns_names,
+                                 ca=ca_name,
+                                 profile=profile,
+                                 msg='Certificate request id {key} with '
+                                     'profile {profile} for CA {ca} does not '
+                                     'have a DNS SAN {san} matching name '
+                                     '{hostname}')
+            if not error:
+                yield Result(self, constants.SUCCESS,
+                             key=request_id,
+                             hostname=hostlist,
+                             san=cert.san_a_label_dns_names,
+                             ca=ca_name,
+                             profile=profile)
+
+
 @registry
 class IPACertNSSTrust(IPAPlugin):
     """Compare the NSS trust for the CA certs to a known good value"""
@@ -427,8 +533,11 @@ class IPACertNSSTrust(IPAPlugin):
                 try:
                     expected = expected_trust[nickname]
                 except KeyError:
-                    logger.debug("%s not found in %s, assuming 3rd party"
-                                 % (nickname, paths.PKI_TOMCAT_ALIAS_DIR))
+                    logger.debug(
+                        "%s not found in %s, assuming 3rd party",
+                        nickname,
+                        paths.PKI_TOMCAT_ALIAS_DIR,
+                    )
                     continue
             try:
                 expected_trust.pop(nickname)
@@ -444,8 +553,8 @@ class IPACertNSSTrust(IPAPlugin):
                     msg='Incorrect NSS trust for {nickname} in {dbdir}. '
                         'Got {got} expected {expected}.')
                 continue
-            else:
-                yield Result(self, constants.SUCCESS, key=nickname)
+
+            yield Result(self, constants.SUCCESS, key=nickname)
 
         for nickname in expected_trust:
             yield Result(
@@ -550,7 +659,7 @@ class IPANSSChainValidation(IPAPlugin):
                                      key=key)
         finally:
             if ca_pw_fname:
-                installutils.remove_file(ca_pw_fname)
+                ipautil.remove_file(ca_pw_fname)
 
 
 @registry
@@ -775,7 +884,7 @@ class IPACertRevocation(IPAPlugin):
                 logger.debug('\'%s\' was not issued by IPA, skipping',
                              DN(cert.subject))
                 continue
-            elif issued is None:
+            if issued is None:
                 logger.debug('LDAP is down, skipping \'%s\'',
                              DN(cert.subject))
                 continue
@@ -866,8 +975,9 @@ class IPACAChainExpirationCheck(IPAPlugin):
                          msg='Error opening IPA CA chain at {key}: {error}')
             return
         except ValueError as e:
-            logger.debug("% contains an invalid certificate" %
-                         paths.IPA_CA_CRT)
+            logger.debug(
+                "% contains an invalid certificate", paths.IPA_CA_CRT
+            )
             yield Result(self, constants.ERROR,
                          key=paths.IPA_CA_CRT,
                          error=str(e),
@@ -876,19 +986,19 @@ class IPACAChainExpirationCheck(IPAPlugin):
             return
 
         now = datetime.now(timezone.utc)
-        soon = now + timedelta(days=self.config.cert_expiration_days)
+        soon = now + timedelta(days=int(self.config.cert_expiration_days))
         for cert in ca_certs:
             subject = DN(cert.subject)
             subject = str(subject).replace('\\;', '\\3b')
             dt = cert.not_valid_after.replace(tzinfo=timezone.utc)
             if dt < now:
-                logger.debug("%s is expired" % subject)
+                logger.debug("%s is expired", subject)
                 yield Result(self, constants.CRITICAL,
                              path=paths.IPA_CA_CRT,
                              key=subject,
                              msg='CA \'{key}\' in {path} is expired.')
             elif dt <= soon:
-                logger.debug("%s is expiring soon" % subject)
+                logger.debug("%s is expiring soon", subject)
                 yield Result(self, constants.WARNING,
                              path=paths.IPA_CA_CRT,
                              key=subject,
