@@ -1,6 +1,7 @@
 #
 # Copyright (C) 2019 FreeIPA Contributors see COPYING for license
 #
+import re
 
 from dns import (
     rdata,
@@ -8,8 +9,9 @@ from dns import (
     rdatatype,
     message,
     rrset,
+    version,
 )
-from dns.resolver import Answer, NoAnswer
+from dns.resolver import Answer
 
 from base import BaseTest
 from util import capture_results, m_api
@@ -25,10 +27,19 @@ from ipaserver.dns_data_management import (
     IPA_DEFAULT_ADTRUST_SRV_REC
 )
 
+try:
+    # pylint: disable=unused-import
+    from ipaserver.install.installutils import resolve_rrsets_nss  # noqa: F401
+    # pylint: enable=unused-import
+except ImportError:
+    resolve_rrsets_import = 'ipaserver.dns_data_management.resolve_rrsets'
+else:
+    resolve_rrsets_import = 'ipaserver.install.installutils.resolve_rrsets_nss'
+
 
 def add_srv_records(qname, port_map, priority=0, weight=100):
     rdlist = []
-    for name, port in port_map:
+    for _, port in port_map:
         answerlist = []
         for host in qname:
             hostname = DNSName(host)
@@ -43,6 +54,19 @@ def add_srv_records(qname, port_map, priority=0, weight=100):
     return rdlist
 
 
+def resolve_rrsets(fqdn, rdtypes):
+    """
+    Return an A record for the hostname in an RRset type in a list.
+    """
+    rset = []
+    for rdtype in rdtypes:
+        rlist = rrset.from_text_list(fqdn, 86400, rdataclass.IN,
+                                     rdtype, gen_addrs(rdtype, 1))
+        rset.append(rlist)
+
+    return rset
+
+
 def query_srv(qname, ad_records=False):
     """
     Return a SRV for each service IPA cares about for all the hosts.
@@ -55,25 +79,16 @@ def query_srv(qname, ad_records=False):
     return rdlist
 
 
-class Response:
-    """Fake class so that a DNS NoAnswer can be raised"""
-    def __init__(self, question=None):
-        self.question = question
-
-    @property
-    def question(self):
-        return self.__question
-
-    @question.setter
-    def question(self, question):
-        self.__question = question
-
-
-def gen_addrs(num):
+def gen_addrs(rdtype=rdatatype.A, num=1):
     """Generate sequential IP addresses for the ipa-ca A record lookup"""
     ips = []
+    if rdtype == rdatatype.A:
+        ip_template = '192.168.0.%d'
+    if rdtype == rdatatype.AAAA:
+        ip_template = '2001:db8:1::%d'
+
     for i in range(num):
-        ips.append('192.168.0.%d' % (i + 1))
+        ips.append(ip_template % (i + 1))
 
     return ips
 
@@ -83,30 +98,39 @@ def fake_query(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN, count=1,
     """Fake a DNS query, returning count responses to the request
 
        Three kinds of lookups are faked:
-       1. A query for A records for a service will return the count
+       1. A query for A/AAAA records for a service will return the count
           as requested in the test. This simulates lookups for the
-          ipa-ca A record. To force a difference in responses one can
+          ipa-ca A/AAAA record. To force a difference in responses one can
           vary the count.
-       2. AAAA records are not yet supported, return no answer
-       3. TXT queries will return the Kerberos realm
+       2. TXT queries will return the Kerberos realm
 
        fake_txt will set an invalid Kerberos realm entry to provoke a
        warning.
     """
-    m = message.Message()
-    if rdtype == rdatatype.A:
+    if version.MAJOR < 2 or (version.MAJOR == 2 and version.MINOR == 0):
+        m = message.Message()
+    elif version.MAJOR == 2 and version.MINOR > 0:
+        m = message.QueryMessage()  # pylint: disable=E1101
+        m = message.make_response(m)  # pylint: disable=E1101
+    if rdtype in (rdatatype.A, rdatatype.AAAA):
         fqdn = DNSName(qname)
         fqdn = fqdn.make_absolute()
 
-        answers = Answer(fqdn, rdataclass.IN, rdatatype.A, m,
-                         raise_on_no_answer=False)
+        if version.MAJOR < 2:
+            # pylint: disable=unexpected-keyword-arg
+            answers = Answer(fqdn, rdataclass.IN, rdtype, m,
+                             raise_on_no_answer=False)
+            # pylint: enable=unexpected-keyword-arg
+        else:
+            if version.MAJOR == 2 and version.MINOR > 0:
+                question = rrset.RRset(fqdn, rdataclass.IN, rdtype)
+                m.question = [question]
+            answers = Answer(fqdn, rdataclass.IN, rdtype, m)
 
         rlist = rrset.from_text_list(fqdn, 86400, rdataclass.IN,
-                                     rdatatype.A, gen_addrs(count))
+                                     rdtype, gen_addrs(rdtype, count))
 
         answers.rrset = rlist
-    elif rdtype == rdatatype.AAAA:
-        raise NoAnswer(response=Response('no AAAA'))
     elif rdtype == rdatatype.TXT:
         if fake_txt:
             realm = 'FAKE_REALM'
@@ -115,8 +139,16 @@ def fake_query(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN, count=1,
         qname = DNSName('_kerberos.' + m_api.env.domain)
         qname = qname.make_absolute()
 
-        answers = Answer(qname, rdataclass.IN, rdatatype.TXT, m,
-                         raise_on_no_answer=False)
+        if version.MAJOR < 2:
+            # pylint: disable=unexpected-keyword-arg
+            answers = Answer(qname, rdataclass.IN, rdatatype.TXT, m,
+                             raise_on_no_answer=False)
+            # pylint: enable=unexpected-keyword-arg
+        else:
+            if version.MAJOR == 2 and version.MINOR > 0:
+                question = rrset.RRset(qname, rdataclass.IN, rdtype)
+                m.question = [question]
+            answers = Answer(qname, rdataclass.IN, rdatatype.TXT, m)
 
         rlist = rrset.from_text_list(qname, 86400, rdataclass.IN,
                                      rdatatype.TXT, [realm])
@@ -173,12 +205,16 @@ class TestDNSSystemRecords(BaseTest):
        2. fake_query() overrides dns.resolver.query to simulate
           A, AAAA and TXT record lookups.
     """
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_single(self, mock_query, mock_query_srv):
+    def test_dnsrecords_single(self, mock_query, mock_query_srv, mock_rrset):
         """Test single CA master, all SRV records"""
         mock_query.side_effect = fake_query_one
         mock_query_srv.side_effect = query_srv([m_api.env.host])
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -197,22 +233,28 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 9
+        assert len(self.results) == 10
 
         for result in self.results.results:
             assert result.result == constants.SUCCESS
             assert result.source == 'ipahealthcheck.ipa.idns'
             assert result.check == 'IPADNSSystemRecordsCheck'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_two(self, mock_query, mock_query_srv):
+    def test_dnsrecords_two(self, mock_query, mock_query_srv, mock_rrset):
         """Test two CA masters, all SRV records"""
         mock_query_srv.side_effect = query_srv([
             m_api.env.host,
             'replica.' + m_api.env.domain
         ])
         mock_query.side_effect = fake_query_two
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -239,16 +281,17 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 17
+        assert len(self.results) == 19
 
         for result in self.results.results:
             assert result.result == constants.SUCCESS
             assert result.source == 'ipahealthcheck.ipa.idns'
             assert result.check == 'IPADNSSystemRecordsCheck'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_three(self, mock_query, mock_query_srv):
+    def test_dnsrecords_three(self, mock_query, mock_query_srv, mock_rrset):
         """Test three CA masters, all SRV records"""
         mock_query_srv.side_effect = query_srv([
             m_api.env.host,
@@ -256,6 +299,13 @@ class TestDNSSystemRecords(BaseTest):
             'replica2.' + m_api.env.domain
         ])
         mock_query.side_effect = fake_query_three
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica2.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -289,16 +339,18 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 25
+        assert len(self.results) == 28
 
         for result in self.results.results:
             assert result.result == constants.SUCCESS
             assert result.source == 'ipahealthcheck.ipa.idns'
             assert result.check == 'IPADNSSystemRecordsCheck'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_three_mixed(self, mock_query, mock_query_srv):
+    def test_dnsrecords_three_mixed(self, mock_query, mock_query_srv,
+                                    mock_rrset):
         """Test three masters, only one with a CA, all SRV records"""
         mock_query_srv.side_effect = query_srv([
             m_api.env.host,
@@ -306,6 +358,13 @@ class TestDNSSystemRecords(BaseTest):
             'replica2.' + m_api.env.domain
         ])
         mock_query.side_effect = fake_query_one
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica2.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA))
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -337,15 +396,17 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 23
+        assert len(self.results) == 24
 
         for result in self.results.results:
             assert result.result == constants.SUCCESS
             assert result.source == 'ipahealthcheck.ipa.idns'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_missing_server(self, mock_query, mock_query_srv):
+    def test_dnsrecords_missing_server(self, mock_query, mock_query_srv,
+                                       mock_rrset):
         """Drop one of the masters from query_srv
 
            This will simulate missing SRV records and cause a number of
@@ -357,6 +418,13 @@ class TestDNSSystemRecords(BaseTest):
             # replica2 is missing
         ])
         mock_query.side_effect = fake_query_three
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica2.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -390,19 +458,21 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 25
+        assert len(self.results) == 28
 
         ok = get_results_by_severity(self.results.results, constants.SUCCESS)
         warn = get_results_by_severity(self.results.results, constants.WARNING)
-        assert len(ok) == 18
+        assert len(ok) == 21
         assert len(warn) == 7
 
         for result in warn:
             assert result.kw.get('msg') == 'Expected SRV record missing'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_missing_ipa_ca(self, mock_query, mock_query_srv):
+    def test_dnsrecords_missing_ipa_ca(self, mock_query, mock_query_srv,
+                                       mock_rrset):
         """Drop one of the masters from query_srv
 
            This will simulate missing SRV records and cause a number of
@@ -414,6 +484,13 @@ class TestDNSSystemRecords(BaseTest):
             'replica2.' + m_api.env.domain
         ])
         mock_query.side_effect = fake_query_two
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica2.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA))
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -447,22 +524,26 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 25
+        assert len(self.results) == 28
 
         ok = get_results_by_severity(self.results.results, constants.SUCCESS)
         warn = get_results_by_severity(self.results.results, constants.WARNING)
-        assert len(ok) == 24
-        assert len(warn) == 1
+        assert len(ok) == 26
+        assert len(warn) == 2
 
         for result in warn:
-            assert result.kw.get('msg') == \
-                'Got {count} ipa-ca A records, expected {expected}'
+            assert re.match(
+                r'^Got {count} ipa-ca (A|AAAA) records, expected {expected}$',
+                result.kw.get('msg')
+            )
             assert result.kw.get('count') == 2
             assert result.kw.get('expected') == 3
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_extra_srv(self, mock_query, mock_query_srv):
+    def test_dnsrecords_extra_srv(self, mock_query, mock_query_srv,
+                                  mock_rrset):
         """An extra SRV record set exists, report it.
 
            Add an extra master to the query_srv() which will generate
@@ -475,6 +556,15 @@ class TestDNSSystemRecords(BaseTest):
             'replica3.' + m_api.env.domain
         ])
         mock_query.side_effect = fake_query_three
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica2.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+            resolve_rrsets('replica3.' + m_api.env.domain,
+                           (rdatatype.A, rdatatype.AAAA)),
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -508,23 +598,28 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 32
+        assert len(self.results) == 35
 
         ok = get_results_by_severity(self.results.results, constants.SUCCESS)
         warn = get_results_by_severity(self.results.results, constants.WARNING)
-        assert len(ok) == 25
+        assert len(ok) == 28
         assert len(warn) == 7
 
         for result in warn:
             assert result.kw.get('msg') == \
                 'Unexpected SRV entry in DNS'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_bad_realm(self, mock_query, mock_query_srv):
+    def test_dnsrecords_bad_realm(self, mock_query, mock_query_srv,
+                                  mock_rrset):
         """Unexpected Kerberos TXT record"""
         mock_query.side_effect = fake_query_one_txt
         mock_query_srv.side_effect = query_srv([m_api.env.host])
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -543,22 +638,27 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 9
+        assert len(self.results) == 10
 
         ok = get_results_by_severity(self.results.results, constants.SUCCESS)
         warn = get_results_by_severity(self.results.results, constants.WARNING)
-        assert len(ok) == 8
+        assert len(ok) == 9
         assert len(warn) == 1
 
         result = warn[0]
         assert result.kw.get('msg') == 'expected realm missing'
         assert result.kw.get('key') == '\"FAKE_REALM\"'
 
+    @patch(resolve_rrsets_import)
     @patch('ipapython.dnsutil.query_srv')
     @patch('dns.resolver.query')
-    def test_dnsrecords_one_with_ad(self, mock_query, mock_query_srv):
+    def test_dnsrecords_one_with_ad(self, mock_query, mock_query_srv,
+                                    mock_rrset):
         mock_query.side_effect = fake_query_one
         mock_query_srv.side_effect = query_srv([m_api.env.host], True)
+        mock_rrset.side_effect = [
+            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
+        ]
 
         m_api.Command.server_find.side_effect = [{
             'result': [
@@ -578,7 +678,7 @@ class TestDNSSystemRecords(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 15
+        assert len(self.results) == 16
 
         for result in self.results.results:
             assert result.result == constants.SUCCESS
