@@ -21,6 +21,7 @@ from ipahealthcheck.ipa.trust import (IPATrustAgentCheck,
                                       IPATrustControllerPrincipalCheck,
                                       IPATrustControllerServiceCheck,
                                       IPATrustControllerGroupSIDCheck,
+                                      IPATrustControllerAdminSIDCheck,
                                       IPATrustControllerConfCheck,
                                       IPATrustPackageCheck)
 
@@ -66,11 +67,63 @@ class mock_ldap_conn:
         if option == OPT_X_SASL_SSF_MIN:
             return 256
 
+        return None
+
     def search_s(self, base, scope, filterstr=None,
                  attrlist=None, attrsonly=0):
         return tuple()
 
 
+#
+# Construct a setup with two direct trusts and one sub domain
+#
+def trust_find():
+    return [{
+        'result': [
+            {
+                'cn': ['ad.example'],
+            },
+            {
+                'cn': ['child.example'],
+            },
+        ]
+    }]
+
+
+def trustdomain_find():
+    return [
+        {
+            "result": [
+                {
+                    "cn": ["ad.example"],
+                    "ipantflatname": ["ADROOT"],
+                    "ipanttrusteddomainsid": ["S-1-5-21-abc"],
+                    "ipanttrusttype": ["2"],
+                    "ipanttrustattributes": ["8"],
+                },
+                {
+                    "cn": ["child.ad.example"],
+                    "ipantflatname": ["CHILD.ADROOT"],
+                    "ipanttrusteddomainsid": ["S-1-5-22-def"],
+                    "ipanttrusttype": ["2"],
+                    "ipanttrustattributes": ["1"],
+                },
+            ],
+        },
+        {
+            "result": [
+                {
+                    "cn": ["child.example"],
+                    "ipantflatname": ["CHILD"],
+                    "ipanttrusteddomainsid": ["S-1-5-21-ghi"],
+                    "ipanttrusttype": ["2"],
+                    "ipanttrustattributes": ["8"],
+                },
+            ],
+        },
+    ]
+
+
 class SSSDDomain:
     def __init__(self, return_ipa_server_mode=True, provider='ipa'):
         self.return_ipa_server_mode = return_ipa_server_mode
@@ -80,11 +133,13 @@ class SSSDDomain:
         if option in ('id_provider', 'auth_provider', 'chpass_provider',
                       'access_provider'):
             return self.provider
-        elif option == 'ipa_server_mode':
+        if option == 'ipa_server_mode':
             if self.return_ipa_server_mode is None:
                 raise NoOptionError()
             return self.return_ipa_server_mode
 
+        return None
+
 
 class SSSDConfig():
     def __init__(self, return_domains=True, return_ipa_server_mode=True,
@@ -245,31 +300,17 @@ class TestTrustDomains(BaseTest):
         dlresult.returncode = 0
         dlresult.error_log = ''
         dlresult.output = 'implicit_files\nipa.example\nad.example\n' \
-            'child.example\n'
+            'child.ad.example\nchild.example\n'
         olresult = namedtuple('run', ['returncode', 'error_log'])
         olresult.returncode = 0
         olresult.error_log = ''
         olresult.output = 'Online status: Online\n\n'
 
-        mock_run.side_effect = [dlresult, olresult, olresult]
+        mock_run.side_effect = [dlresult, olresult, olresult, olresult]
 
         # get_trust_domains()
-        m_api.Command.trust_find.side_effect = [{
-            'result': [
-                {
-                    'cn': ['ad.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-abc'],
-                    'trusttype': ['Active Directory domain'],
-                },
-                {
-                    'cn': ['child.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-def'],
-                    'trusttype': ['Active Directory domain'],
-                },
-            ]
-        }]
+        m_api.Command.trust_find.side_effect = trust_find()
+        m_api.Command.trustdomain_find.side_effect = trustdomain_find()
 
         framework = object()
         registry.initialize(framework, config.Config)
@@ -278,15 +319,17 @@ class TestTrustDomains(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 3
+        assert len(self.results) == 4
 
         result = self.results.results[0]
         assert result.result == constants.SUCCESS
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustDomainsCheck'
         assert result.kw.get('key') == 'domain-list'
-        assert result.kw.get('trust_domains') == 'ad.example, child.example'
-        assert result.kw.get('sssd_domains') == 'ad.example, child.example'
+        assert result.kw.get('trust_domains') == \
+            'ad.example, child.ad.example, child.example'
+        assert result.kw.get('sssd_domains') == \
+            'ad.example, child.ad.example, child.example'
 
         result = self.results.results[1]
         assert result.result == constants.SUCCESS
@@ -300,6 +343,13 @@ class TestTrustDomains(BaseTest):
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustDomainsCheck'
         assert result.kw.get('key') == 'domain-status'
+        assert result.kw.get('domain') == 'child.ad.example'
+
+        result = self.results.results[3]
+        assert result.result == constants.SUCCESS
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustDomainsCheck'
+        assert result.kw.get('key') == 'domain-status'
         assert result.kw.get('domain') == 'child.example'
 
     @patch('ipapython.ipautil.run')
@@ -318,22 +368,8 @@ class TestTrustDomains(BaseTest):
         mock_run.side_effect = [dlresult, olresult, olresult]
 
         # get_trust_domains()
-        m_api.Command.trust_find.side_effect = [{
-            'result': [
-                {
-                    'cn': ['ad.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-abc'],
-                    'trusttype': ['Active Directory domain'],
-                },
-                {
-                    'cn': ['child.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-def'],
-                    'trusttype': ['Active Directory domain'],
-                },
-            ]
-        }]
+        m_api.Command.trust_find.side_effect = trust_find()
+        m_api.Command.trustdomain_find.side_effect = trustdomain_find()
 
         framework = object()
         registry.initialize(framework, config.Config)
@@ -349,7 +385,8 @@ class TestTrustDomains(BaseTest):
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustDomainsCheck'
         assert result.kw.get('key') == 'domain-list'
-        assert result.kw.get('trust_domains') == 'ad.example, child.example'
+        assert result.kw.get('trust_domains') == \
+            'ad.example, child.ad.example, child.example'
         assert result.kw.get('sssd_domains') == 'child.example'
 
         result = self.results.results[1]
@@ -427,29 +464,16 @@ class TestTrustCatalog(BaseTest):
         ds2result.output = 'Active servers:\nAD Global Catalog: ' \
             'root-dc.ad.vm\nAD Domain Controller: root-dc.ad.vm\n' \
 
-        mock_run.side_effect = [dsresult, ds2result]
+        mock_run.side_effect = [dsresult, dsresult, ds2result]
         mock_getnamebysid.side_effect = [
            {'S-1-5-21-abc-500': {'name': 'admin@ad.example', 'type': 3}},
+           {'S-1-5-21-ghi-500': {'name': 'admin@child.ad.example', 'type': 3}},
            {'S-1-5-21-def-500': {'name': 'admin@child.example', 'type': 3}}
         ]
 
         # get_trust_domains()
-        m_api.Command.trust_find.side_effect = [{
-            'result': [
-                {
-                    'cn': ['ad.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-abc'],
-                    'trusttype': ['Active Directory domain'],
-                },
-                {
-                    'cn': ['child.example'],
-                    'ipantflatname': ['ADROOT'],
-                    'ipanttrusteddomainsid': ['S-1-5-21-def'],
-                    'trusttype': ['Active Directory domain'],
-                },
-            ]
-        }]
+        m_api.Command.trust_find.side_effect = trust_find()
+        m_api.Command.trustdomain_find.side_effect = trustdomain_find()
 
         framework = object()
         registry.initialize(framework, config.Config)
@@ -458,7 +482,7 @@ class TestTrustCatalog(BaseTest):
 
         self.results = capture_results(f)
 
-        assert len(self.results) == 6
+        assert len(self.results) == 9
 
         result = self.results.results[0]
         assert result.result == constants.SUCCESS
@@ -486,20 +510,40 @@ class TestTrustCatalog(BaseTest):
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustCatalogCheck'
         assert result.kw.get('key') == 'Domain Security Identifier'
-        assert result.kw.get('sid') == 'S-1-5-21-def'
+        assert result.kw.get('sid') == 'S-1-5-22-def'
 
         result = self.results.results[4]
         assert result.result == constants.SUCCESS
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustCatalogCheck'
         assert result.kw.get('key') == 'AD Global Catalog'
-        assert result.kw.get('domain') == 'child.example'
+        assert result.kw.get('domain') == 'child.ad.example'
 
         result = self.results.results[5]
         assert result.result == constants.SUCCESS
         assert result.source == 'ipahealthcheck.ipa.trust'
         assert result.check == 'IPATrustCatalogCheck'
         assert result.kw.get('key') == 'AD Domain Controller'
+
+        result = self.results.results[6]
+        assert result.result == constants.SUCCESS
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustCatalogCheck'
+        assert result.kw.get('key') == 'Domain Security Identifier'
+        assert result.kw.get('sid') == 'S-1-5-21-ghi'
+
+        result = self.results.results[7]
+        assert result.result == constants.SUCCESS
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustCatalogCheck'
+        assert result.kw.get('key') == 'AD Global Catalog'
+        assert result.kw.get('domain') == 'child.example'
+
+        result = self.results.results[8]
+        assert result.result == constants.SUCCESS
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustCatalogCheck'
+        assert result.kw.get('key') == 'AD Domain Controller'
         assert result.kw.get('domain') == 'child.example'
 
 
@@ -878,6 +922,80 @@ class TestControllerGroupSID(BaseTest):
         assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-500'
 
 
+class TestControllerAdminSID(BaseTest):
+    patches = {
+        'ldap.initialize':
+        Mock(return_value=mock_ldap_conn()),
+    }
+
+    def test_not_trust_controller(self):
+        framework = object()
+        registry.initialize(framework, config.Config)
+        registry.trust_controller = False
+        f = IPATrustControllerAdminSIDCheck(registry)
+
+        self.results = capture_results(f)
+
+        # Zero because the call was skipped altogether
+        assert len(self.results) == 0
+
+    def test_principal_ok(self):
+        admin_dn = DN(('uid', 'admin'))
+        attrs = {
+            'ipantsecurityidentifier':
+            ['S-1-5-21-1234-5678-1976041503-500'],
+        }
+        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
+        ldapentry = LDAPEntry(fake_conn, admin_dn)
+        for attr, values in attrs.items():
+            ldapentry[attr] = values
+
+        framework = object()
+        registry.initialize(framework, config.Config)
+        registry.trust_controller = True
+        f = IPATrustControllerAdminSIDCheck(registry)
+
+        f.conn = mock_ldap(ldapentry)
+        self.results = capture_results(f)
+
+        assert len(self.results) == 1
+
+        result = self.results.results[0]
+        assert result.result == constants.SUCCESS
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustControllerAdminSIDCheck'
+        assert result.kw.get('key') == 'ipantsecurityidentifier'
+        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-500'
+
+    def test_principal_fail(self):
+        admin_dn = DN(('uid', 'admin'))
+        attrs = {
+            'ipantsecurityidentifier':
+            ['S-1-5-21-1234-5678-1976041503-400'],
+        }
+        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
+        ldapentry = LDAPEntry(fake_conn, admin_dn)
+        for attr, values in attrs.items():
+            ldapentry[attr] = values
+
+        framework = object()
+        registry.initialize(framework, config.Config)
+        registry.trust_controller = True
+        f = IPATrustControllerAdminSIDCheck(registry)
+
+        f.conn = mock_ldap(ldapentry)
+        self.results = capture_results(f)
+
+        assert len(self.results) == 1
+
+        result = self.results.results[0]
+        assert result.result == constants.ERROR
+        assert result.source == 'ipahealthcheck.ipa.trust'
+        assert result.check == 'IPATrustControllerAdminSIDCheck'
+        assert result.kw.get('key') == 'ipantsecurityidentifier'
+        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-400'
+
+
 class TestControllerConf(BaseTest):
     patches = {
         'ldap.initialize':
